#include <windows.h>
#include <shlwapi.h>
#include <stdio.h>
#include <conio.h>
#include <errno.h>
#include <intrin.h>
#include "whirlpool\whirlpool.h"
#include "aes\aes.h"
#include "serpent\serpent.h"
#include "twofish\twofish.h"
#include "xts\xts.h"
#include "scrypt\scrypt.h"
#include "misc\misc.h"

#define IMDPROXY_FLAG_RO 0x1
#define IMDPROXY_HEADER_SIZE 4096
#define DEF_BUFFER_SIZE (1 << 20)
#define AIM_BUFFER_SIZE (2 << 20)
#define DEF_REQUIRED_ALIGNMENT 1

typedef enum {
	IMDPROXY_REQ_NULL, IMDPROXY_REQ_INFO,
	IMDPROXY_REQ_READ, IMDPROXY_REQ_WRITE,
	IMDPROXY_REQ_CONNECT, IMDPROXY_REQ_CLOSE
} IMDPROXY_REQ;
typedef struct {ULONGLONG file_size, req_alignment, flags;} IMDPROXY_INFO_RESP;
typedef struct {ULONGLONG request_code, offset, length;} IMDPROXY_IO_REQ;
typedef struct {ULONGLONG errorno, length;} IMDPROXY_IO_RESP;

static HANDLE shm_server_mutex, shm_request_event, shm_response_event;
static unsigned char *shm_view, *main_buf;
static int buffer_size = DEF_BUFFER_SIZE;
static __int64 offset = 0, extra_offset = 0;
static int sector_size = 512;
static __int64 current_size;
static IMDPROXY_INFO_RESP proxy_info = {0};
static TCHAR file_name[32768], backup_file[32768];
static TCHAR drive_letter = 0;
static int verbose = 1;
static char digit_group_str[64];
static TCHAR password[256];
static unsigned int pass_size;
static unsigned char header_key_whirlpool[6][32];
static unsigned char header_key_sha3[6][32];
static unsigned char salt_whirlpool[64 + 4], salt_sha3[64 + 4];
static BOOL detach_console = FALSE;
static aes_key key_aes1, key_aes2;
static serpent_key key_serpent1, key_serpent2;
static twofish_key key_twofish1, key_twofish2;
static int cipher = 0;
static SYSTEM_INFO si;
static HANDLE encrypt_event[32], decrypt_event[32];
static HANDLE encrypt_done[32], decrypt_done[32];
static int request_size_read, request_size_write;
static void *thread_buffer_read, *thread_buffer_write;
static __int64 thread_offset_read, thread_offset_write;
static HANDLE read_event, write_event;
static OVERLAPPED overlap_read, overlap_write = {0};
static unsigned char *write_buf;
static int expected_write = 0;
static BOOL force_aim = FALSE, use_imdisk;
static HANDLE h_cpl;

int cipher_list[CIPHER_LAST + 1][4] = {
	{0},
	{1, CIPHER_AES},
	{1, CIPHER_SERPENT},
	{1, CIPHER_TWOFISH},
	{2, CIPHER_AES, CIPHER_TWOFISH},
	{2, CIPHER_SERPENT, CIPHER_AES},
	{2, CIPHER_SERPENT, CIPHER_TWOFISH},
	{3, CIPHER_SERPENT, CIPHER_AES, CIPHER_TWOFISH}
};
static char *cipher_name[] = {"", "AES", "Serpent", "Twofish", "AES-Twofish", "Serpent-AES", "Serpent-Twofish", "Serpent-AES-Twofish"};

#pragma pack(push, 1)
typedef struct {unsigned char salt[64]; short version;
		unsigned char hash[WHIRLPOOL_DIGESTSIZE]; unsigned char master_key[6][32];
		__int64 volume_offset; __int64 volume_size;
} HEADER_STRUCT;
#pragma pack(pop)

static HANDLE h;


void out(char *format, ...)
{
	if (!verbose) return;
	va_list args;
	va_start(args, format);
	if (detach_console) AttachConsole(ATTACH_PARENT_PROCESS);
	vprintf(format, args);
	if (detach_console) FreeConsole();
	va_end(args);
}

void wout(TCHAR *format, ...)
{
	if (!verbose) return;
	va_list args;
	va_start(args, format);
	if (detach_console) AttachConsole(ATTACH_PARENT_PROCESS);
	vwprintf(format, args);
	if (detach_console) FreeConsole();
	va_end(args);
}

/*void disp(void *buf, int size)
{
	int i;

	for (i = 0; i < size; i++) out("%02X", ((unsigned char*)buf)[i]);
	out("\n");
}*/

DWORD __stdcall derive_pass_whirlpool(LPVOID lpParam)
{
	int retval;

	if ((retval = scrypt(PRF_WHIRLPOOL, (unsigned char*)password, pass_size, salt_whirlpool, 64, 32768, 16, 16, (unsigned char*)header_key_whirlpool, sizeof header_key_whirlpool)))
		out("Cannot allocate memory for scrypt.\n");
	return retval;
}

DWORD __stdcall derive_pass_sha3(LPVOID lpParam)
{
	int retval;

	if ((retval = scrypt(PRF_SHA3, (unsigned char*)password, pass_size, salt_sha3, 64, 32768, 16, 16, (unsigned char*)header_key_sha3, sizeof header_key_sha3)))
		out("Cannot allocate memory for scrypt.\n");
	return retval;
}


DWORD __stdcall encrypt_thread(LPVOID lpParam)
{
	int i, i_start = (size_t)lpParam << 9, i_inc = si.dwNumberOfProcessors << 9;

	for (;;) {
		WaitForSingleObject(encrypt_event[(int)(size_t)lpParam], INFINITE);
		for (i = i_start; i < request_size_write; i += i_inc)
			XTS_EncryptSector(cipher, (thread_offset_write + i) >> 9, thread_buffer_write + i);
		SetEvent(encrypt_done[(int)(size_t)lpParam]);
	}
}

DWORD __stdcall decrypt_thread(LPVOID lpParam)
{
	int i, i_start = (size_t)lpParam << 9, i_inc = si.dwNumberOfProcessors << 9;

	for (;;) {
		WaitForSingleObject(decrypt_event[(int)(size_t)lpParam], INFINITE);
		for (i = i_start; i < request_size_read; i += i_inc)
			XTS_DecryptSector(cipher, (thread_offset_read + i) >> 9, thread_buffer_read + i);
		SetEvent(decrypt_done[(int)(size_t)lpParam]);
	}
}


int physical_read(void *buf, int size, __int64 offset)
{
	int i, current_size, next_size, size_done = 0;
	__int64 next_offset;
	DWORD bytes_read;

	next_size = min(size, 32768);
	overlap_read.Internal = overlap_read.InternalHigh = 0;
	overlap_read.Offset = offset;
	overlap_read.OffsetHigh = offset >> 32;
	overlap_read.hEvent = read_event;
	ReadFile(h, buf, next_size, NULL, &overlap_read);

	do {
		current_size = next_size;
		next_size = min(size - (size_done + current_size), 65536);
		GetOverlappedResult(h, &overlap_read, &bytes_read, TRUE);
		if (bytes_read != current_size) {
			out("%Read error: %u / %d bytes at offset %I64d.\n", bytes_read, current_size, ((__int64)overlap_read.OffsetHigh << 32) + overlap_read.Offset);
			return EIO;
		}
		if (next_size) {
			overlap_read.Internal = overlap_read.InternalHigh = 0;
			next_offset = offset + size_done + current_size;
			overlap_read.Offset = next_offset;
			overlap_read.OffsetHigh = next_offset >> 32;
			overlap_read.hEvent = read_event;
			ReadFile(h, buf + size_done + current_size, next_size, NULL, &overlap_read);
		}
		thread_buffer_read = buf + size_done;
		request_size_read = current_size;
		thread_offset_read = offset + size_done;
		for (i = 0; i < si.dwNumberOfProcessors; i++)
			SetEvent(decrypt_event[i]);
		WaitForMultipleObjects(si.dwNumberOfProcessors, decrypt_done, TRUE, INFINITE);
		size_done += current_size;
	} while (next_size);

	if (verbose >= 2) out(" ? Read: %d bytes at offset %I64d\n", size, offset);

	return 0;
}

int physical_write(void *buf, int size, __int64 offset)
{
	int i, err = 0;
	DWORD bytes_written;

	thread_buffer_write = buf;
	request_size_write = size;
	thread_offset_write = offset;
	for (i = 0; i < si.dwNumberOfProcessors; i++)
		SetEvent(encrypt_event[i]);
	WaitForMultipleObjects(si.dwNumberOfProcessors, encrypt_done, TRUE, INFINITE);

	GetOverlappedResult(h, &overlap_write, &bytes_written, TRUE);
	if (bytes_written != expected_write) {
		out("%Write error: %u / %d bytes at offset %I64d.\n", bytes_written, expected_write, ((__int64)overlap_write.OffsetHigh << 32) + overlap_write.Offset);
		err = EIO;
	}
	memcpy(write_buf, buf, size);
	overlap_write.Internal = overlap_write.InternalHigh = 0;
	overlap_write.Offset = offset;
	overlap_write.OffsetHigh = offset >> 32;
	overlap_write.hEvent = write_event;
	WriteFile(h, write_buf, size, NULL, &overlap_write);
	expected_write = size;
	if (verbose >= 2) out(" ! Write: %d bytes at offset %I64d\n", size, offset);

	return err;
}


int do_comm()
{
	IMDPROXY_IO_REQ *req_block;
	IMDPROXY_IO_RESP *resp_block;
	HANDLE hFileMap;
	ULARGE_INTEGER map_size;
	TCHAR comm_device[20];
	TCHAR objname[120], *index;
	STARTUPINFO si = {sizeof si};
	PROCESS_INFORMATION pi;

	swprintf(comm_device, L"ProxyCrypt%x", GetTickCount());
	index = objname + swprintf(objname, L"Global\\%s", comm_device);

	map_size.QuadPart = buffer_size + IMDPROXY_HEADER_SIZE;

	if (!(hFileMap = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE | SEC_COMMIT, map_size.HighPart, map_size.LowPart, objname))) {
		out("CreateFileMapping() failed.\nPlease check that you have administrative privileges.\n");
		return 2;
	}
	if (!(shm_view = MapViewOfFile(hFileMap, FILE_MAP_WRITE, 0, 0, 0))) {
		out("MapViewOfFile() failed.\n");
		return 2;
	}

	main_buf = shm_view + IMDPROXY_HEADER_SIZE;
	req_block = (IMDPROXY_IO_REQ*)shm_view;
	resp_block = (IMDPROXY_IO_RESP*)shm_view;

	wcscpy(index, L"_Server");
	if (!(shm_server_mutex = CreateMutex(NULL, FALSE, objname))) {
		out("CreateMutex() failed.\n");
		return 2;
	}
	if (WaitForSingleObject(shm_server_mutex, 0) != WAIT_OBJECT_0) {
		out("A service with this name is already running.\n");
		return 2;
	}

	wcscpy(index, L"_Request");
	if (!(shm_request_event = CreateEvent(NULL, FALSE, FALSE, objname))) {
		out("CreateEvent() failed.\n");
		return 2;
	}
	wcscpy(index, L"_Response");
	if (!(shm_response_event = CreateEvent(NULL, FALSE, FALSE, objname))) {
		out("CreateEvent() failed.\n");
		return 2;
	}

	if (use_imdisk) swprintf(objname, L"imdisk.exe -a -t proxy -m %c: -o shm -f %s -S %d", drive_letter, comm_device, sector_size);
	else swprintf(objname, L"ArsenalImageMounterControl.exe /add /filename=%s /sectorsize=%d /proxy=shm", comm_device, sector_size);
	CreateProcess(NULL, objname, NULL, NULL, FALSE, verbose ? 0 : CREATE_NO_WINDOW, NULL, NULL, &si, &pi);
	FreeConsole();
	detach_console = TRUE;

	if (WaitForSingleObject(shm_request_event, use_imdisk ? 10000 : 20000) != WAIT_OBJECT_0) {
		out("WaitForSingleObject() failed.\n");
		return 2;
	}
	if (req_block->request_code != IMDPROXY_REQ_INFO) {
		out("Incorrect first request.\n");
		return 2;
	}

	wout(L"Connection on object %s.\n", comm_device);

	memcpy(shm_view, &proxy_info, sizeof proxy_info);

	for (;;) {
		if (!SetEvent(shm_response_event)) {
			out("SetEvent() failed.\n");
			return 1;
		}
		if (WaitForSingleObject(shm_request_event, INFINITE) != WAIT_OBJECT_0) {
			out("Error while flushing comm data.\n");
			return 1;
		}

		if (req_block->request_code == IMDPROXY_REQ_READ)
			resp_block->errorno = physical_read(main_buf, req_block->length, offset + req_block->offset);
		else if (req_block->request_code == IMDPROXY_REQ_WRITE)
			resp_block->errorno = physical_write(main_buf, req_block->length, offset + req_block->offset);
		else {
			out("Connection closed.\n");
			return 0;
		}

		resp_block->length = req_block->length;
	}
}

#pragma GCC optimize "Os"

LRESULT __stdcall WndProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	HANDLE device;
	char path[40];
	STARTUPINFOA si;
	PROCESS_INFORMATION pi;
	DWORD dw;

	if (Msg == WM_ENDSESSION) {
		sprintf(path, "\\\\.\\%c:", (char)drive_letter);
		if (use_imdisk) {
			if ((device = CreateFileA(path, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_NO_BUFFERING, NULL)) == INVALID_HANDLE_VALUE)
				device = CreateFileA(path, 0, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_NO_BUFFERING, NULL);
			else
				FlushFileBuffers(device);
			if (!DeviceIoControl(device, FSCTL_DISMOUNT_VOLUME, NULL, 0, NULL, 0, &dw, NULL) || !DeviceIoControl(device, IOCTL_STORAGE_EJECT_MEDIA, NULL, 0, NULL, 0, &dw, NULL))
				GetProcAddress(h_cpl, "ImDiskForceRemoveDevice")(device, 0);
			CloseHandle(device);
		} else {
			for (path[4] = 'A'; path[4] <= 'Z'; path[4]++)
				if ((device = CreateFileA(path, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_NO_BUFFERING, NULL)) != INVALID_HANDLE_VALUE) {
					FlushFileBuffers(device);
					CloseHandle(device);
				}
			ZeroMemory(&si, sizeof si);
			si.cb = sizeof si;
			strcpy(path, "ArsenalImageMounterControl.exe /remove");
			CreateProcessA(NULL, path, NULL, NULL, FALSE, CREATE_NO_WINDOW, NULL, NULL, &si, &pi);
			WaitForSingleObject(pi.hProcess, INFINITE);
		}
		return 0;
	} else
		return DefWindowProc(hWnd, Msg, wParam, lParam);
}

DWORD __stdcall msg_window(LPVOID lpParam)
{
	MSG msg;
	WNDCLASSA wc = {0};

	wc.lpfnWndProc = WndProc;
	wc.lpszClassName = "X";
	RegisterClassA(&wc);
	CreateWindowA("X", NULL, 0, 0, 0, 0, 0, NULL, NULL, NULL, NULL);

	for (;;) {
		GetMessage(&msg, NULL, 0, 0);
		DispatchMessage(&msg);
	}
}

void init_key(unsigned char key[6][32])
{
	int i, j;

	for (i = 0; i < cipher_list[cipher][0]; i++) {
		j = i * 2;
		switch (cipher_list[cipher][i + 1]) {
			case CIPHER_AES:
				memcpy(key_aes1.key, key[j], 32);
				aes_init(&key_aes1);
				memcpy(key_aes2.key, key[j + 1], 32);
				aes_init(&key_aes2);
				break;
			case CIPHER_SERPENT:
				memcpy(key_serpent1.key, key[j], 32);
				serpent_init(&key_serpent1);
				memcpy(key_serpent2.key, key[j + 1], 32);
				serpent_init(&key_serpent2);
				break;
			case CIPHER_TWOFISH:
				memcpy(key_twofish1.key, key[j], 32);
				twofish_init(&key_twofish1);
				memcpy(key_twofish2.key, key[j + 1], 32);
				twofish_init(&key_twofish2);
		}
	}
	keys[CIPHER_AES][0] = &key_aes1;
	keys[CIPHER_AES][1] = &key_aes2;
	keys[CIPHER_SERPENT][0] = &key_serpent1;
	keys[CIPHER_SERPENT][1] = &key_serpent2;
	keys[CIPHER_TWOFISH][0] = &key_twofish1;
	keys[CIPHER_TWOFISH][1] = &key_twofish2;
}

char *group_digit(unsigned long long n)
{
	int size;

	size = sprintf(digit_group_str, "%I64u", n);
	while (size > 3) {
		size -= 3;
		memmove(digit_group_str + size + 1, digit_group_str + size, 24);
		digit_group_str[size] = ' ';
	}
	return digit_group_str;
}

__declspec(noinline) void process_suffix(__int64 *size, TCHAR suffix)
{
	switch (suffix) {
		case 't': case 'T': *size <<= 10;
		case 'g': case 'G': *size <<= 10;
		case 'm': case 'M': *size <<= 10;
		case 'k': case 'K': *size <<= 10;
	}
}

void format_size(__int64 size, int drive_number, __int64 offset, VOLUME_DISK_EXTENTS *vde, char *result)
{
	int i;
	char tmp[20];

	if (size < 10000) sprintf(result, "%6I64d", size);
	else if (size >> 10 < 10000) sprintf(result, "%6I64d KB", size >> 10);
	else if (size >> 20 < 10000) sprintf(result, "%6I64d MB", size >> 20);
	else if (size >> 30 < 10000) sprintf(result, "%6I64d GB", size >> 30);
	else sprintf(result, "%6I64d TB", size >> 40);

	for (i = 0; i < 26; i++)
		if (vde[i].Extents[0].DiskNumber == drive_number && vde[i].Extents[0].StartingOffset.QuadPart == offset) {
			strcpy(tmp, result);
			sprintf(result, "%s  (%c:)", tmp, 'A' + i);
		}
}

void display_drives()
{
	int drive_number;
	int partition_number, ext_number;
	int entry_size;
	unsigned char *mbr;
	unsigned char partition_type;
	DISK_GEOMETRY_EX dge;
	DWORD n_bytes;
	__int64 drive_size, partition_size;
	__int64 offset_ext;
	char txt_size[20];
	int i;
	VOLUME_DISK_EXTENTS vde[26];

	for (i = 0; i < 26; i++) {
		swprintf(file_name, L"\\\\.\\%c:", 'A' + i);
		h = CreateFile(file_name, 0, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (!DeviceIoControl(h, IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS, NULL, 0, &vde[i], sizeof(VOLUME_DISK_EXTENTS), &n_bytes, NULL))
			vde[i].Extents[0].DiskNumber = -1;
		CloseHandle(h);
	}

	mbr = VirtualAlloc(NULL, 34 * 4096, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);
	for (drive_number = 0; ; drive_number++) {
		swprintf(file_name, L"\\\\.\\PhysicalDrive%d", drive_number);
		h = CreateFile(file_name, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (h == INVALID_HANDLE_VALUE) break;
		sector_size = 512;
		drive_size = 0;
		if (DeviceIoControl(h, IOCTL_DISK_GET_DRIVE_GEOMETRY_EX, NULL, 0, &dge, sizeof dge, &n_bytes, NULL)) {
			sector_size = dge.Geometry.BytesPerSector;
			drive_size = dge.DiskSize.QuadPart;
		}
		if (sector_size > 4096) continue;
		format_size(drive_size, drive_number, -1, vde, txt_size);
		printf("Drive %d:%s\n", drive_number, txt_size);
		physical_read(mbr, 2 * 4096, 0);
		if (mbr[0x01BE + 4] == 0xEE && !memcmp(mbr + sector_size, "EFI PART\0\0\1", 12)) {
			entry_size = *(int*)(mbr + sector_size + 84);
			physical_read(mbr, 34 * sector_size, 0);
			for (partition_number = 0; partition_number < 128; partition_number++) {
				i = ((partition_number >> 2) + 2) * sector_size + (partition_number & 3) * entry_size;
				if (!(*(__int64*)(mbr + i) | *(__int64*)(mbr + i + 8))) break;
				offset = *(__int64*)(mbr + i + 32) * sector_size;
				partition_size = (*(__int64*)(mbr + i + 40) + 1) * sector_size - offset;
				format_size(partition_size, drive_number, offset, vde, txt_size);
				printf(" -partition %d:%s\n", partition_number + 1, txt_size);
			}
		}
		else if (*(USHORT*)(mbr + 0x01FE) == 0xAA55 && ((mbr[0x01BE] | mbr[0x01CE] | mbr[0x01DE] | mbr[0x01EE]) & 0x7F) == 0) {
			for (partition_number = 1; partition_number <= 4; partition_number++) {
				partition_size = (__int64)(*(UINT*)(mbr + 0x01AE + partition_number * 16 + 12)) * sector_size;
				if (!partition_size) break;
				format_size(partition_size, drive_number, (__int64)(*(UINT*)(mbr + 0x01AE + partition_number * 16 + 8)) * sector_size, vde, txt_size);
				printf(" -partition %d:%s\n", partition_number, txt_size);
				partition_type = mbr[0x01AE + partition_number * 16 + 4];
				if (partition_type == 0x05 || partition_type == 0x0F) {
					offset_ext = offset = (__int64)(*(UINT*)(mbr + 0x01AE + partition_number * 16 + 8)) * sector_size;
					ext_number = 5;
					do {
						physical_read(mbr, 4096, offset);
						partition_size = (__int64)(*(UINT*)(mbr + 0x01BE + 12)) * sector_size;
						format_size(partition_size, drive_number, offset + (__int64)(*(UINT*)(mbr + 0x01BE + 8)) * sector_size, vde, txt_size);
						printf(" --partition %d:%s\n", ext_number++, txt_size);
						offset = offset_ext + (__int64)(*(UINT*)(mbr + 0x01CE + 8)) * sector_size;
					} while (offset != offset_ext);
				}
			}
		}
		CloseHandle(h);
		printf("\n");
	}
	VirtualFree(mbr, 0, MEM_RELEASE);
}

void benchmark()
{
	HCRYPTPROV hProv;
	DWORD init_time, enc_time, dec_time, n_enc, n_dec;
	int i;

	thread_buffer_read = thread_buffer_write = VirtualAlloc(NULL, DEF_BUFFER_SIZE, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);
	request_size_read = request_size_write = DEF_BUFFER_SIZE;
	thread_offset_read = thread_offset_write = 0;
	whirlpool_ini();
	if (!CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, 0))
		CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, CRYPT_NEWKEYSET);
	CryptGenRandom(hProv, DEF_BUFFER_SIZE, thread_buffer_read);
	CryptReleaseContext(hProv, 0);
	printf("The following tests are made in RAM in XTS mode.\n\nCIPHER\t\t      ENCRYPT   DECRYPT   AVERAGE (MB/s)\n\n");
	for (cipher = 1; cipher <= CIPHER_LAST; cipher++) {
		init_key((void*)thread_buffer_read);
		n_enc = n_dec = 0;
		init_time = GetTickCount();
		do {
			for (i = 0; i < si.dwNumberOfProcessors; i++)
				SetEvent(encrypt_event[i]);
			WaitForMultipleObjects(si.dwNumberOfProcessors, encrypt_done, TRUE, INFINITE);
			enc_time = GetTickCount();
			n_enc++;
		} while (enc_time - init_time < 500);
		do {
			for (i = 0; i < si.dwNumberOfProcessors; i++)
				SetEvent(decrypt_event[i]);
			WaitForMultipleObjects(si.dwNumberOfProcessors, decrypt_done, TRUE, INFINITE);
			dec_time = GetTickCount();
			n_dec++;
		} while (dec_time - enc_time < 500);
		printf("%-19s%10u%10u%10u\n", cipher_name[cipher],
			   (UINT)((__int64)n_enc * DEF_BUFFER_SIZE * 1000 / (enc_time - init_time) >> 20),
			   (UINT)((__int64)n_dec * DEF_BUFFER_SIZE * 1000 / (dec_time - enc_time) >> 20),
			   (UINT)((__int64)(n_enc + n_dec) * DEF_BUFFER_SIZE * 1000 / (dec_time - init_time) >> 20));
	}
}

int wmain(int argc, TCHAR *argv[])
{
	int drive_number = -1;
	int partition_number = 1;
	BOOL force_partition = FALSE;
	int entry_size;
	unsigned char *mbr;
	__int64 offset_ext;
	BOOL new_volume = FALSE, change_pass = FALSE, bench = FALSE;
	__int64 new_file_size = 0;
	int n_hash = 0;
	__int64 max_volume_size, size_new_volume = 0;
	DISK_GEOMETRY_EX dge;
	DWORD n_bytes;
	TCHAR suffix;
	TCHAR path[MAX_PATH];
	int i, j, retval = 0;
	int new_cipher;
	HEADER_STRUCT *header;
	union {unsigned char ch[512 + 64]; HEADER_STRUCT header; __m128i i128;} tmp_header;
	unsigned char *buf, hash[WHIRLPOOL_DIGESTSIZE];
	unsigned char header_key[6][32];
	HANDLE h_backup = NULL;
	HANDLE thread_handle[2];
	DWORD init_time = GetTickCount();
	int CPUInfo[4];
	BOOL mount = FALSE;
	FILETIME access_time, write_time;
	DWORD drive_mask;

	if (argc < 2) {
		printf("ProxyCrypt 1.1.2\n\n"
			"ProxyCrypt [drive:] -f file | -d drivenumber[,[partition]] [-c [size]]\n"
			"           [-p partition] [-o offset] [-cp] [-r] [-t threads] [-b backup_file]\n"
			"           [-v verbose] [-aim]\n"
			"ProxyCrypt -l\n"
			"ProxyCrypt -bm [-t threads]\n\n"
			" -f\tImage file of the encrypted volume.\n\n"
			" -c\tCreate a new encrypted volume. If a size and a file name are specified,\n"
			"\tan image file of this size will be created. Use k, m, g or t as suffix\n"
			"\tto specify KB, MB, GB or TB.\n\n"
			" -p\tNumber of partition (default: 1). Use 0 to ignore the partition table\n"
			"\t(if any) and to use the whole disk or image.\n\n"
			" -d\tNumber of physical drive where the encrypted volume is located.\n"
			"\tIf a comma is added without partition number, the latter will be\n"
			"\tassumed to be 0 instead of 1.\n\n"
			" -o\tOffset to the encrypted volume inside the file/drive/partition.\n"
			"\tUse k, m, g, or t as suffix to specify KB, MB, GB or TB.\n\n"
			" -cp\tChange password.\n\n"
			" -r\tMount a volume as read-only.\n\n"
			" -t\tNumber of threads used for decryption and encryption, from 1 to 32\n"
			"\t(the default value is the number of logical processors).\n\n"
			" -b\tIf backup_file does not exist, it will be created with the content of\n"
			"\tthe volume header, encrypted with a new salt.\n"
			"\tIf backup_file exists, the volume header will be overwritten with the\n"
			"\tfirst 512 bytes of backup_file.\n\n"
			" -v\tVerbosity level, from 0 to 2 (default: 1).\n\n"
			" -aim\tForce the use of Arsenal Image Mounter, in case of the ImDisk Virtual\n"
			"\tDisk Driver, which is used by default, is also installed.\n"
			"\tIf Arsenal Image Mounter is used, drive: is ignored.\n\n"
			" -l\tDisplay a list of the current drives and partitions.\n\n"
			" -bm\tBenchmark of available ciphers.\n");
		return 0;
	}

	GetSystemInfo(&si);
	__cpuid(CPUInfo, 1);
	if (!(CPUInfo[3] & 0x4000000)) {
		printf("SSE2 instructions required.\n");
		return 1;
	}

	if (!_wcsicmp(argv[1], L"-l")) {
		display_drives();
		return 0;
	}

	argc--;
	file_name[0] = backup_file[0] = 0;
	do {
		argv++;
		if (!_wcsicmp(argv[0], L"-bm")) {
			bench = TRUE;
			continue;
		}
		if (argv[0][1] == ':' && argv[0][2] == 0) {
			if (argv[0][0] < 'A' || (argv[0][0] > 'Z' && argv[0][0] < 'a') || argv[0][0] > 'z') {
syntax_error:
				printf("Syntax error.\n");
				return 1;
			}
			if (PathFileExists(argv[0])) {
				wprintf(L"%s already in use.\n", argv[0]);
				return 1;
			}
			drive_letter = argv[0][0];
			continue;
		}
		if (!_wcsicmp(argv[0], L"-f")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			if (wcslen(argv[0]) < MAX_PATH)
				wcscpy(file_name, argv[0]);
			else {
				_snwprintf(file_name, _countof(file_name) - 1, L"\\\\?\\%s", argv[0]);
				file_name[_countof(file_name) - 1] = 0;
			}
			continue;
		}
		if (!_wcsicmp(argv[0], L"-c")) {
			new_volume = TRUE;
			if (argc > 1) {
				suffix = 0;
				if (swscanf(argv[1], L"%I64d%c", &new_file_size, &suffix)) {
					process_suffix(&new_file_size, suffix);
					if (new_file_size < 28 * 1024) {
						printf("New file size must be at least 28 KB.\n");
						return 1;
					}
					argc--;
					argv++;
				}
			}
			continue;
		}
		if (!_wcsicmp(argv[0], L"-p")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			swscanf(argv[0], L"%u", &partition_number);
			force_partition = TRUE;
			continue;
		}
		if (!_wcsicmp(argv[0], L"-d")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			switch (swscanf(argv[0], L"%u%c%u", &drive_number, &suffix, &partition_number)) {
				case 2: partition_number = 0;
				case 3: force_partition = TRUE;
			}
			continue;
		}
		if (!_wcsicmp(argv[0], L"-o")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			suffix = 0;
			swscanf(argv[0], L"%I64d%c", &extra_offset, &suffix);
			process_suffix(&extra_offset, suffix);
			continue;
		}
		if (!_wcsicmp(argv[0], L"-cp")) {
			change_pass = TRUE;
			continue;
		}
		if (!_wcsicmp(argv[0], L"-r")) {
			proxy_info.flags = IMDPROXY_FLAG_RO;
			continue;
		}
		if (!_wcsicmp(argv[0], L"-t")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			swscanf(argv[0], L"%u", &si.dwNumberOfProcessors);
			continue;
		}
		if (!_wcsicmp(argv[0], L"-b")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			if (wcslen(argv[0]) < MAX_PATH)
				wcscpy(backup_file, argv[0]);
			else {
				_snwprintf(backup_file, _countof(backup_file) - 1, L"\\\\?\\%s", argv[0]);
				backup_file[_countof(backup_file) - 1] = 0;
			}
			continue;
		}
		if (!_wcsicmp(argv[0], L"-v")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			verbose = *argv[0] - '0';
			continue;
		}
		if (!_wcsicmp(argv[0], L"-aim"))
			force_aim = TRUE;
	} while (--argc > 0);

	if (CPUInfo[2] & 0x2000000) {
		out("AES instructions detected.\n");
		aes_encrypt = aes_encrypt_hw;
		aes_decrypt = aes_decrypt_hw;
	} else {
		aes_encrypt = aes_encrypt_sw;
		aes_decrypt = aes_decrypt_sw;
	}

	// Create threads
	if (si.dwNumberOfProcessors < 1) si.dwNumberOfProcessors = 1;
	if (si.dwNumberOfProcessors > 32) si.dwNumberOfProcessors = 32;
	out("Uses %u thread%s for encryption/decryption.\n", si.dwNumberOfProcessors, si.dwNumberOfProcessors > 1 ? "s" : "");
	for (i = 0; i < si.dwNumberOfProcessors; i++) {
		if (!(encrypt_event[i] = CreateEvent(NULL, FALSE, FALSE, NULL)) || !(decrypt_event[i] = CreateEvent(NULL, FALSE, FALSE, NULL)) ||
			!(encrypt_done[i] = CreateEvent(NULL, FALSE, FALSE, NULL)) || !(decrypt_done[i] = CreateEvent(NULL, FALSE, FALSE, NULL))) {
			out("CreateEvent() failed.\n");
			return 1;
		}
		if (!CreateThread(NULL, 0, encrypt_thread, (LPVOID)(size_t)i, 0, NULL) || !CreateThread(NULL, 0, decrypt_thread, (LPVOID)(size_t)i, 0, NULL)) {
			out("CreateThread() failed.\n");
			return 1;
		}
	}

	read_event = CreateEvent(NULL, FALSE, FALSE, NULL);
	write_event = CreateEvent(NULL, FALSE, TRUE, NULL);

	if (bench) {
		benchmark();
		return 0;
	}

	wcscpy(path, L"imdisk.exe");
	use_imdisk = PathFileExists(path) | PathFindOnPath(path, NULL);
	if (!use_imdisk || force_aim) {
		wcscpy(path, L"ArsenalImageMounterControl.exe");
		if (PathFileExists(path) || PathFindOnPath(path, NULL)) {
			use_imdisk = FALSE;
			buffer_size = AIM_BUFFER_SIZE;
		} else {
			out(force_aim ? "Cannot find ArsenalImageMounterControl.exe.\n" : "Can find neither imdisk.exe nor ArsenalImageMounterControl.exe.\n");
			return 1;
		}
	}
	if (use_imdisk) {
		if (!drive_letter) {
			drive_mask = GetLogicalDrives() >> 3;
			for (drive_letter = 'D'; drive_letter <= 'Z'; drive_letter++, drive_mask >>= 1)
				if ((drive_mask & 1) == 0) break;
			if (drive_letter > 'Z') {
				out("Cannot find free drive letter.\n");
				return 1;
			}
		}
		if (!(h_cpl = LoadLibraryA("imdisk.cpl"))) {
			out("Cannot find imdisk.cpl.\n");
			return 1;
		}
	}

	write_buf = VirtualAlloc(NULL, buffer_size, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);

	if (drive_number != -1)
		swprintf(file_name, L"\\\\.\\PhysicalDrive%d", drive_number);
	else if (new_file_size) {
		// Creating file
		if (proxy_info.flags) {
			out("Cannot create file in read-only mode.\n");
			return 1;
		}
		if ((h = CreateFile(file_name, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_NEW, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED | FILE_FLAG_WRITE_THROUGH, NULL)) == INVALID_HANDLE_VALUE) {
			out(GetLastError() == ERROR_FILE_EXISTS ? "Error: file already exists.\n" : "Error: file cannot be created.\n");
			return 1;
		}
		GetFileTime(h, NULL, &access_time, &write_time);
		SetFilePointerEx(h, (LARGE_INTEGER)new_file_size, NULL, FILE_BEGIN);
		if (!SetEndOfFile(h)) {
			CloseHandle(h);
			DeleteFile(file_name);
			out("Cannot create a file of the specified size.\n");
			return 1;
		}
		SetFilePointerEx(h, (LARGE_INTEGER)0LL, NULL, FILE_BEGIN);
		proxy_info.file_size = new_file_size;
	}

	if (!new_file_size) {
		// Opening file or drive
		if (drive_number == -1) {
			if ((h = CreateFile(file_name, proxy_info.flags ? GENERIC_READ : GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING,
								FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED | FILE_FLAG_WRITE_THROUGH, NULL)) == INVALID_HANDLE_VALUE) {
				wout(L"Cannot open %s.\n", file_name);
				return 1;
			}
			GetFileTime(h, NULL, &access_time, &write_time);
		} else {
			if ((h = CreateFile(file_name, proxy_info.flags ? GENERIC_READ : GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
								FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED | FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH, NULL)) == INVALID_HANDLE_VALUE) {
				out("Cannot open drive %d.\n", drive_number);
				return 1;
			}
		}
		GetFileSizeEx(h, (LARGE_INTEGER*)&proxy_info.file_size);
	}

	if (drive_number != -1) {
		// Getting sector size
		if (DeviceIoControl(h, IOCTL_DISK_GET_DRIVE_GEOMETRY_EX, NULL, 0, &dge, sizeof dge, &n_bytes, NULL)) {
			sector_size = dge.Geometry.BytesPerSector;
			proxy_info.file_size = dge.DiskSize.QuadPart;
		}
		if (sector_size < 512) {
			out("Sector size must be at least 512-bytes (detected: %d)\n", sector_size);
			return 1;
		}
		if (sector_size > 4096) {
			out("Sector size must be at most 4096-bytes (detected: %d)\n", sector_size);
			return 1;
		}
	}

	if (!proxy_info.file_size)
		out("Cannot determine size of image/drive.\n");
	current_size = proxy_info.file_size;

	header = VirtualAlloc(NULL, 4096 * 2, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);
	buf = VirtualAlloc(NULL, 4096 * 2, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);

	// Getting offset to the partition
	if (partition_number >= 1 && partition_number <= 128 && current_size >= 2 * sector_size) {
		mbr = VirtualAlloc(NULL, 34 * sector_size, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);
		physical_read(mbr, 2 * 4096, 0);

		if (mbr[0x01BE + 4] == 0xEE && !memcmp(mbr + sector_size, "EFI PART\0\0\1", 12)) {
			out("GPT detected\n");

			entry_size = *(int*)(mbr + sector_size + 84);
			physical_read(mbr, 34 * sector_size, 0);
			i = (((partition_number - 1) >> 2) + 2) * sector_size + ((partition_number - 1) & 3) * entry_size;
			if (*(__int64*)(mbr + i) | *(__int64*)(mbr + i + 8)) {
				offset = *(__int64*)(mbr + i + 32) * sector_size;
				proxy_info.file_size = (*(__int64*)(mbr + i + 40) + 1) * sector_size - offset;
			} else if (force_partition) {
				out("Partition %i not found.\n", partition_number);
				goto clean;
			}
		}
		else {
			if (*(USHORT*)(mbr + 0x01FE) == 0xAA55 && ((mbr[0x01BE] | mbr[0x01CE] | mbr[0x01DE] | mbr[0x01EE]) & 0x7F) == 0) {
				out("MBR detected.\n");

				if (partition_number >= 5) {
					for (i = 0x01BE; i < 0x01FE; i += 16)
						if (mbr[i + 4] == 0x05 || mbr[i + 4] == 0x0F) {
							offset_ext = offset = (__int64)(*(UINT*)(mbr + i + 8)) * sector_size;
							physical_read(mbr, 4096, offset);
							for (j = 5; j < partition_number; j++) {
								offset = offset_ext + (__int64)(*(UINT*)(mbr + 0x01CE + 8)) * sector_size;
								if (offset == offset_ext) {
									out("Partition %i not found.\n", partition_number);
									goto clean;
								}
								physical_read(mbr, 4096, offset);
							}
							offset += (__int64)(*(UINT*)(mbr + 0x01BE + 8)) * sector_size;
							proxy_info.file_size = (__int64)(*(UINT*)(mbr + 0x01BE + 12)) * sector_size;
							break;
						}
					if (i == 0x01FE) {
						out("No extended partition found.\n", partition_number);
						goto clean;
					}
				} else {
					offset = (__int64)(*(UINT*)(mbr + 0x01AE + partition_number * 16 + 8)) * sector_size;
					proxy_info.file_size = (__int64)(*(UINT*)(mbr + 0x01AE + partition_number * 16 + 12)) * sector_size;
				}

				if (proxy_info.file_size)
					out("Partition %i used.\n", partition_number);
				else {
					if (force_partition) {
						out("Partition %i not found.\n", partition_number);
						goto clean;
					} else {
						offset = 0;
						proxy_info.file_size = current_size;
					}
				}
			} else
				out("No MBR/GPT detected. Uses entire image.\n");
		}
		VirtualFree(mbr, 0, MEM_RELEASE);
	}

	proxy_info.req_alignment = DEF_REQUIRED_ALIGNMENT;
	out("Total size: %s bytes.\n", group_digit(current_size));
	out("%s bytes used from offset ", group_digit(proxy_info.file_size));
	out("%s.\n", group_digit(offset));

	offset += extra_offset;
	if (proxy_info.file_size < 20992) {
		out("Too small volume.\n");
		goto clean;
	}
	if (proxy_info.file_size - extra_offset < 20992) {
		out("Offset to the encrypted volume too high.\n");
		goto clean;
	}

	if (!VirtualLock(password, sizeof password) |
		!VirtualLock(header_key_whirlpool, sizeof header_key_whirlpool) |
		!VirtualLock(header_key_sha3, sizeof header_key_sha3) |
		!VirtualLock(&key_aes1, sizeof key_aes1) |
		!VirtualLock(&key_aes2, sizeof key_aes2) |
		!VirtualLock(&key_serpent1, sizeof key_serpent1) |
		!VirtualLock(&key_serpent2, sizeof key_serpent2) |
		!VirtualLock(&key_twofish1, sizeof key_twofish1) |
		!VirtualLock(&key_twofish2, sizeof key_twofish2) |
		!VirtualLock(&tmp_header, sizeof tmp_header) |
		!VirtualLock(header, 4096 * 2) |
		!VirtualLock(buf, 4096 * 2))
		out("Warning: cannot lock data into physical memory.\n");

	whirlpool_ini();

	if (new_volume) {
		printf("\n-Select encryption algorithm:\n1: AES\n2: Serpent\n3: Twofish\n4: AES-Twofish\n5: Serpent-AES\n6: Serpent-Twofish\n7: Serpent-AES-Twofish\n");
		fwscanf(stdin, L"%d", &new_cipher);
		if (new_cipher < 1 || new_cipher > CIPHER_LAST) {
			printf("Unrecognized input. AES selected by default.\n");
			new_cipher = 1;
		}

		printf("\n-Select password hash algorithm:\n1: Whirlpool\n2: SHA-3-512\n");
		fwscanf(stdin, L"%d", &n_hash);
		if (n_hash < 1 || n_hash > 2) {
			printf("Unrecognized input. Whirlpool selected by default.\n");
			n_hash = 1;
		}

		tmp_header.header.volume_offset = (extra_offset + 512 + 4095) & ~4095;
		max_volume_size = proxy_info.file_size - tmp_header.header.volume_offset;
		printf("\n-Enter size of the encrypted volume (28 KB - %I64u KB).\n Use k, m, g, or t as suffix, 0 for the maximum available.\n", max_volume_size >> 10);
		suffix = 0;
		fwscanf(stdin, L"%I64d%c", &size_new_volume, &suffix);
		process_suffix(&size_new_volume, suffix);
		if (!size_new_volume) size_new_volume = max_volume_size;
		if (size_new_volume < 28 * 1024) size_new_volume = 28 * 1024;
		if (size_new_volume > max_volume_size) size_new_volume = max_volume_size;
		tmp_header.header.volume_size = size_new_volume;

		for (;;) {
			out("\nEnter password (press Tab to show/hide):\n");
			if ((retval = password_edit(password)))
				goto clean;
			wcscpy((TCHAR*)buf, password);
			out("Confirm password (press Tab to show/hide):\n");
			if ((retval = password_edit(password)))
				goto clean;
			if (!wcscmp(password, (TCHAR*)buf)) break;
			out("Passwords are different. Please retry.\n");
		}

		out("Creating master keys...\n");
		if (rand_gen(buf, 512 + 4096, n_hash - 1, init_time)) goto clean;
		memcpy(tmp_header.header.salt, buf, 64);
		memcpy(tmp_header.header.master_key, buf + 64, sizeof header->master_key);
		whirlpool((unsigned char*)tmp_header.header.master_key, sizeof header->master_key, tmp_header.header.hash, 1, 1);

		out("Hashing password...\n");
		if (scrypt(n_hash - 1, (unsigned char*)password, wcslen(password) * sizeof(TCHAR), buf, 64, 32768, 16, 16, (unsigned char*)header_key, sizeof header_key)) {
			out("Cannot allocate memory for scrypt.\n");
			goto clean;
		}
		tmp_header.header.version = 1;

		physical_read(header, 2 * 4096, offset & ~4095);
		cipher = new_cipher;
		init_key(header_key);
		XTS_EncryptSector(cipher, offset, tmp_header.ch + 64);
		cipher = 0;
		i = offset % 4096;
		memcpy((unsigned char*)header + i, tmp_header.ch, 512);
		memcpy((unsigned char*)header + i + 512, buf + 512, 4095 - (i + 511) % 4096);
		physical_write(header, 2 * 4096, offset & ~4095);
		out("Volume created successfully.\n");
		XTS_DecryptSector(cipher = new_cipher, offset, tmp_header.ch + 64);
	}
	else
	{
		physical_read(buf, 2 * 4096, offset & ~4095);
		memcpy(header, (char*)buf + (offset & 4095), 512);

		if (backup_file[0]) {
			if ((h_backup = CreateFile(backup_file, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE) {
				wout(L"Cannot open %s.\n", backup_file);
				goto clean;
			}
			retval = GetLastError();
			if (retval == ERROR_ALREADY_EXISTS) {
				if (!ReadFile(h_backup, (char*)header, 512, &n_bytes, NULL) || n_bytes != 512) {
					out("Error while reading backup file.\n");
					goto clean;
				}
			}
		}

		out("\nEnter password (press Tab to show/hide):\n");
		if (password_edit(password))
			goto clean;

		out("Checking password...");
		pass_size = wcslen(password) * sizeof(TCHAR);
		memcpy(salt_whirlpool, header->salt, 64);
		memcpy(salt_sha3, header->salt, 64);
		thread_handle[PRF_WHIRLPOOL] = CreateThread(NULL, 0, derive_pass_whirlpool, NULL, 0, NULL);
		thread_handle[PRF_SHA3] = CreateThread(NULL, 0, derive_pass_sha3, NULL, 0, NULL);
		if (WaitForMultipleObjects(_countof(thread_handle), thread_handle, TRUE, INFINITE) == WAIT_FAILED) {
			out("\nWaitForMultipleObjects() failed.\n");
			goto clean;
		}
		for (cipher = 1; cipher <= CIPHER_LAST; cipher++) {
			init_key(header_key_whirlpool);
			memcpy(tmp_header.ch, (char*)header, 512);
			XTS_DecryptSector(cipher, offset, tmp_header.ch + 64);
			whirlpool((unsigned char*)tmp_header.header.master_key, sizeof header->master_key, hash, 1, 1);
			if (!memcmp(hash, tmp_header.header.hash, WHIRLPOOL_DIGESTSIZE)) break;
		}
		if (cipher > CIPHER_LAST) {
			n_hash = 1;
			for (cipher = 1; cipher <= CIPHER_LAST; cipher++) {
				init_key(header_key_sha3);
				memcpy(tmp_header.ch, (char*)header, 512);
				XTS_DecryptSector(cipher, offset, tmp_header.ch + 64);
				whirlpool((unsigned char*)tmp_header.header.master_key, sizeof header->master_key, hash, 1, 1);
				if (!memcmp(hash, tmp_header.header.hash, WHIRLPOOL_DIGESTSIZE)) break;
			}
		}
		if (cipher > CIPHER_LAST) {
			out(" Error.\n");
			goto clean;
		}
		out(" OK.\nPassword hashed with scrypt/%s.\nCipher is %s.\n", n_hash ? "SHA-3-512" : "Whirlpool", cipher_name[cipher]);

		if (backup_file[0]) {
			if (!retval) {
				out("Encrypting header with new salt...\n");
				memcpy((char*)header, tmp_header.ch, 512);
				if (rand_gen(buf, 256, n_hash - 1, init_time)) goto clean;
				memcpy(header->salt, buf, 64);
				if (scrypt(n_hash - 1, (unsigned char*)password, wcslen(password) * sizeof(TCHAR), buf, 64, 32768, 16, 16, (unsigned char*)header_key, sizeof header_key)) {
					out("Cannot allocate memory for scrypt.\n");
					goto clean;
				}
				init_key(header_key);
				XTS_EncryptSector(cipher, offset, (unsigned char*)header + 64);
				if (!WriteFile(h_backup, (char*)header, 512, &n_bytes, NULL) || n_bytes != 512) {
					out("Error while saving volume header.\n");
					goto clean;
				}
				out("Volume header saved successfully.\n");
			}
			if (retval == ERROR_ALREADY_EXISTS) {
				memcpy((char*)buf + (offset & 4095), header, 512);
				new_cipher = cipher;
				cipher = 0;
				physical_write(buf, 2 * 4096, offset & ~4095);
				out("Volume header written successfully.\n");
				cipher = new_cipher;
			}
			CloseHandle(h_backup);
		}

		if (change_pass) {
			for (;;) {
				out("\nEnter new password (press Tab to show/hide):\n");
				if ((retval = password_edit(password)))
					goto clean;
				wcscpy((TCHAR*)buf, password);
				out("Confirm new password (press Tab to show/hide):\n");
				if ((retval = password_edit(password)))
					goto clean;
				if (!wcscmp(password, (TCHAR*)buf)) break;
				out("Passwords are different. Please retry.\n");
			}
			out("\n-Select password hash algorithm:\n1: Whirlpool\n2: SHA-3-512\n");
			fwscanf(stdin, L"%d", &n_hash);
			if (n_hash < 1 || n_hash > 2) {
				out("Unrecognized input. Whirlpool selected by default.\n");
				n_hash = 1;
			}
			out("Hashing new password...\n");
			if (rand_gen(buf, 256, n_hash - 1, init_time)) goto clean;
			memcpy(tmp_header.header.salt, buf, 64);
			if (scrypt(n_hash - 1, (unsigned char*)password, wcslen(password) * sizeof(TCHAR), buf, 64, 32768, 16, 16, (unsigned char*)header_key, sizeof header_key)) {
				out("Cannot allocate memory for scrypt.\n");
				goto clean;
			}
			init_key(header_key);
			XTS_EncryptSector(cipher, offset, tmp_header.ch + 64);
			new_cipher = cipher;
			cipher = 0;
			physical_read(header, 2 * 4096, offset & ~4095);
			memcpy((char*)header + offset % 4096, tmp_header.ch, 512);
			physical_write(header, 2 * 4096, offset & ~4095);
			out("Password changed.\n");
			XTS_DecryptSector(cipher = new_cipher, offset, tmp_header.ch + 64);
		}
	}

	init_key(tmp_header.header.master_key);
	offset += tmp_header.header.volume_offset - extra_offset;
	proxy_info.file_size = tmp_header.header.volume_size;
	out("Encrypted volume of %s bytes at offset ", group_digit(proxy_info.file_size));
	out("%s.\n\n", group_digit(offset));
	mount = TRUE;

clean:
	SecureZeroMemory(buf, 4096 * 2);
	SecureZeroMemory(header, 4096 * 2);
	VirtualFree(buf, 0, MEM_RELEASE);
	VirtualFree(header, 0, MEM_RELEASE);
	SecureZeroMemory(password, sizeof password);
	SecureZeroMemory(header_key_whirlpool, sizeof header_key_whirlpool);
	SecureZeroMemory(header_key_sha3, sizeof header_key_sha3);
	SecureZeroMemory(&tmp_header, sizeof tmp_header);
	SecureZeroMemory(header_key, sizeof header_key);
	SecureZeroMemory(hash, sizeof hash);

	if (mount) {
		SetProcessShutdownParameters(0x100, 0);
		CreateThread(NULL, 0, msg_window, NULL, 0, NULL);
		retval = do_comm();
	}

	aes_close(&key_aes1);
	aes_close(&key_aes2);
	serpent_close(&key_serpent1);
	serpent_close(&key_serpent2);
	twofish_close(&key_twofish1);
	twofish_close(&key_twofish2);
	SetFileTime(h, NULL, &access_time, &write_time);
	CloseHandle(h);

	return retval | !mount;
}

#ifdef MINGW_NOSTDLIB
void mainCRTStartup()
{
	int argc;
	LPWSTR *argv = CommandLineToArgvW(GetCommandLine(), &argc);
	ExitProcess(wmain(argc, argv));
}
#endif
