#include <intrin.h>
#include "..\aes\aes.h"
#include "..\serpent\serpent.h"
#include "..\twofish\twofish.h"
#include "xts.h"

#pragma GCC optimize "O2"

#define Mult(T) \
{	__m128i tmp = _mm_shuffle_epi32(_mm_srli_epi64(T.i128, 63), _MM_SHUFFLE(1, 0, 3, 2)); \
	T.i128 = (_mm_slli_epi64(T.i128, 1) | tmp) ^ _mm_mul_epu32(tmp, GF_128_FDBK); }

void XTS_EncryptSector(int cipher, unsigned long long sector, unsigned char *buffer)
{
	void *key1, *key2;
	union {unsigned char ch[64]; __m128i i128;} T;
	union {unsigned char ch[16]; __m128i i128;} T2, T3, T4;
	unsigned int i, n_cipher, n, current_cipher;
	unsigned char *ptr;
	__m128i GF_128_FDBK = _mm_set_epi64x(0, 0x87 - 1);

	n_cipher = cipher_list[cipher][0];
	for (n = 0; n < n_cipher; n++) {
		ptr = buffer;

		current_cipher = cipher_list[cipher][n_cipher - n];
		key1 = keys[current_cipher][0];
		key2 = keys[current_cipher][1];

		T.i128 = _mm_loadl_epi64((__m128i*)&sector);

		switch (current_cipher) {
			case CIPHER_AES:
				aes_encrypt(key2, T.ch);
				for (i = 0; i < 512; i += 16, ptr += 16) {
					*(__m128i*)ptr ^= T.i128;
					aes_encrypt(key1, ptr);
					*(__m128i*)ptr ^= T.i128;
					Mult(T);
				}
				break;

			case CIPHER_SERPENT:
				serpent_encrypt(key2, T.ch);
				for (i = 0; i < 512; i += 64, ptr += 64) {
					T2.i128 = T.i128;
					Mult(T2);
					T3.i128 = T2.i128;
					Mult(T3);
					T4.i128 = T3.i128;
					Mult(T4);
					*(__m128i*)ptr ^= T.i128;
					*(__m128i*)(ptr + 16) ^= T2.i128;
					*(__m128i*)(ptr + 32) ^= T3.i128;
					*(__m128i*)(ptr + 48) ^= T4.i128;
					serpent_encrypt(key1, ptr);
					*(__m128i*)ptr ^= T.i128;
					*(__m128i*)(ptr + 16) ^= T2.i128;
					*(__m128i*)(ptr + 32) ^= T3.i128;
					*(__m128i*)(ptr + 48) ^= T4.i128;
					T.i128 = T4.i128;
					Mult(T);
				}
				break;

			case CIPHER_TWOFISH:
				twofish_encrypt(key2, T.ch);
				for (i = 0; i < 512; i += 16, ptr += 16) {
					*(__m128i*)ptr ^= T.i128;
					twofish_encrypt(key1, ptr);
					*(__m128i*)ptr ^= T.i128;
					Mult(T);
				}
		}
	}
}

void XTS_DecryptSector(int cipher, unsigned long long sector, unsigned char *buffer)
{
	void *key1, *key2;
	union {unsigned char ch[64]; __m128i i128;} T;
	union {unsigned char ch[16]; __m128i i128;} T2, T3, T4;
	unsigned int i, n_cipher, n, current_cipher;
	unsigned char *ptr;
	__m128i GF_128_FDBK = _mm_set_epi64x(0, 0x87 - 1);

	n_cipher = cipher_list[cipher][0];
	for (n = 0; n < n_cipher; n++) {
		ptr = buffer;

		current_cipher = cipher_list[cipher][n + 1];
		key1 = keys[current_cipher][0];
		key2 = keys[current_cipher][1];

		T.i128 = _mm_loadl_epi64((__m128i*)&sector);

		switch (current_cipher) {
			case CIPHER_AES:
				aes_encrypt(key2, T.ch);
				for (i = 0; i < 512; i += 16, ptr += 16) {
					*(__m128i*)ptr ^= T.i128;
					aes_decrypt(key1, ptr);
					*(__m128i*)ptr ^= T.i128;
					Mult(T);
				}
				break;

			case CIPHER_SERPENT:
				serpent_encrypt(key2, T.ch);
				for (i = 0; i < 512; i += 64, ptr += 64) {
					T2.i128 = T.i128;
					Mult(T2);
					T3.i128 = T2.i128;
					Mult(T3);
					T4.i128 = T3.i128;
					Mult(T4);
					*(__m128i*)ptr ^= T.i128;
					*(__m128i*)(ptr + 16) ^= T2.i128;
					*(__m128i*)(ptr + 32) ^= T3.i128;
					*(__m128i*)(ptr + 48) ^= T4.i128;
					serpent_decrypt(key1, ptr);
					*(__m128i*)ptr ^= T.i128;
					*(__m128i*)(ptr + 16) ^= T2.i128;
					*(__m128i*)(ptr + 32) ^= T3.i128;
					*(__m128i*)(ptr + 48) ^= T4.i128;
					T.i128 = T4.i128;
					Mult(T);
				}
				break;

			case CIPHER_TWOFISH:
				twofish_encrypt(key2, T.ch);
				for (i = 0; i < 512; i += 16, ptr += 16) {
					*(__m128i*)ptr ^= T.i128;
					twofish_decrypt(key1, ptr);
					*(__m128i*)ptr ^= T.i128;
					Mult(T);
				}
		}
	}
}
